# generate_workerman_data.py

from collections import Counter
import locale

import natsort
import networkx as nx
from pulp import LpProblem
from tabulate import tabulate

from bdo_empire.generate_graph_data import GraphData


def get_workerman_json(workers, ref_data, lodging):
    """Populate and return a standard 'dummy' instance of the workerman dict."""
    lodgingP2W = {}
    for group in ref_data["groups"]:
        if group in ref_data["group_to_townname"]:
            townname = ref_data["group_to_townname"][group]
            lodgingP2W[group] = lodging[townname]
    workerman_json = {
        "activateAncado": False,
        "lodgingP2W": lodgingP2W,
        "userWorkers": workers,
        "farmingEnable": False,
        "farmingProfit": 0,
        "farmingBareProfit": 0,
        "grindTakenList": [],
    }
    return workerman_json


def make_workerman_worker(town_id: int, origin_id: int, worker_data: dict, stash_id: int):
    """Populate and return a 'dummy' instance of a workerman user worker dict."""
    worker = {
        "tnk": town_id,
        "charkey": str(worker_data["charkey"]),
        "label": "default",
        "level": 40,
        "wspdSheet": worker_data["wspd"],
        "mspdSheet": worker_data["mspd"],
        "luckSheet": worker_data["luck"],
        "skills": [int(s) for s in worker_data["skills"]],
        "job": {"kind": "plantzone", "pzk": int(origin_id), "storage": stash_id},
    }
    return worker


def order_workerman_workers(graph, user_workers: list[dict], solution_distances):
    """Order user workers into import order for correct workerman paths construction."""

    # Order by shortest origin -> town paths to break ties by nearest nodes.
    distance_indices = zip(list(range(len(solution_distances))), solution_distances)
    distance_indices = sorted(distance_indices, key=lambda x: x[1])
    workerman_user_workers = [user_workers[i] for i, _ in distance_indices]

    # Iterative ordering of user workers by shortest paths with weight removal on used arcs.
    ordered_workers = []
    while workerman_user_workers:
        distances = []
        all_pairs = dict(nx.all_pairs_bellman_ford_path_length(graph, weight="weight"))
        for worker in workerman_user_workers:
            distance = all_pairs[str(worker["tnk"])][str(worker["job"]["pzk"])]
            distances.append(distance)
        min_value = min(distances)
        min_indice = distances.index(min_value)
        worker = workerman_user_workers[min_indice]
        ordered_workers.append(worker)
        workerman_user_workers.pop(min_indice)

        short_path = nx.shortest_path(graph, str(worker["tnk"]), str(worker["job"]["pzk"]), "weight")
        for s, d in zip(short_path, short_path[1:]):
            if graph.edges[(s, d)]["weight"] >= 1:
                for edge in graph.in_edges(d):
                    graph.edges[edge]["weight"] = 0
                break

    return ordered_workers


def generate_graph(graph_data: GraphData, prob):
    graph = nx.DiGraph()
    exclude_keywords = ["lodging", "𝓢", "𝓣"]

    for var_key, var in prob.variablesDict().items():
        if not round(var.varValue) >= 1:
            continue
        exclude = any(keyword in var_key for keyword in exclude_keywords)
        if exclude:
            continue

        u, v = None, None
        if "groupflow_" in var_key and "_on_" in var_key:
            tmp = var_key.split("_on_")
            tmp = tmp[1].split("_to_")
            u = tmp[0]
            v = tmp[1]
        else:
            continue

        source, destination = u, v
        weight = graph_data["V"][source].cost
        graph.add_edge(destination.split("_")[1], source.split("_")[1], weight=weight)

    return graph


def extract_solution(prob) -> tuple[dict, dict, dict]:
    lodging_vars = {}
    origin_vars = {}
    waypoint_vars = {}
    for k, v in prob.variablesDict().items():
        if not round(v.varValue) >= 1:
            continue
        if k.startswith("flow_lodging_") and "_to_" not in k:
            lodging_vars[k.replace("flow_", "")] = v
        elif "_on_plant_" in k:
            origin_vars[k.split("_")[4]] = k.split("_")[1]
        elif k.startswith("x_waypoint") or k == "x_town_1343":
            waypoint_vars[k.replace("x_", "")] = v
    return lodging_vars, origin_vars, waypoint_vars


def process_solution(origin_vars: dict, data: dict, graph_data: GraphData, graph: nx.DiGraph):
    all_pairs = dict(nx.all_pairs_bellman_ford_path_length(graph, weight="weight"))

    calculated_value = 0
    distances = []
    origin_cost = 0
    outputs = []
    town_ids = set()
    workerman_user_workers = []
    root_ranks = []
    stash_town_id = 601
    for k, v in origin_vars.items():
        town_id = data["group_to_town"][v]
        town_ids.add(town_id)
        distances.append(all_pairs[town_id][k])

        origin = graph_data["V"][f"plant_{k}"]
        worker_data = origin.group_prizes[v]["worker_data"]
        user_worker = make_workerman_worker(int(town_id), int(origin.id), worker_data, stash_town_id)
        workerman_user_workers.append(user_worker)

        value = origin.group_prizes[v]["value"]
        worker = origin.group_prizes[v]["worker"]
        root_rank = list(origin.group_prizes.keys()).index(v) + 1
        root_ranks.append(root_rank)

        origin_cost += origin.cost
        calculated_value += value

        output = {
            "warehouse": v,
            "node": origin.id,
            "worker": worker,
            "value": locale.currency(round(value), grouping=True, symbol=True)[:-3],
            "value_rank": root_rank,
        }
        outputs.append(output)

    return calculated_value, distances, origin_cost, outputs, workerman_user_workers


def print_summary(outputs, counts: dict, costs: dict, total_value: float):
    """Print town, origin, worker summary report."""
    outputs = natsort.natsorted(outputs, key=lambda x: (x["warehouse"], x["node"]))
    colalign = ("right", "right", "left", "right", "right")
    print(tabulate(outputs, headers="keys", colalign=colalign))
    print("By Town:\n\n", tabulate([[k, v] for k, v in counts["by_groups"].items()]), "\n")
    print("  Lodging cost:", costs["lodgings"])
    print("  Worker Nodes:", counts["origins"], "cost:", costs["origins"])
    print("     Waypoints:", counts["waypoints"], "cost:", costs["waypoints"])
    print("    Total Cost:", sum(c for c in costs.values()))
    print("         Value:", locale.currency(round(total_value), grouping=True, symbol=True)[:-3])


def generate_workerman_data(
    prob: LpProblem, lodging: dict, data: dict, graph_data: GraphData
) -> dict:
    print("Creating workerman json...")
    locale.setlocale(locale.LC_ALL, "")

    graph = generate_graph(graph_data, prob)
    lodging_vars, origin_vars, waypoint_vars = extract_solution(prob)
    solution = process_solution(origin_vars, data, graph_data, graph)
    calculated_value, distances, origin_cost, outputs, workerman_user_workers = solution
    workerman_ordered_workers = order_workerman_workers(graph, workerman_user_workers, distances)
    workerman_json = get_workerman_json(workerman_ordered_workers, data, lodging)

    counts: dict = {"origins": len(origin_vars), "waypoints": len(waypoint_vars)}
    counts["by_groups"] = {
        str(data["group_to_townname"][k]): v for k, v in Counter(origin_vars.values()).most_common()
    }
    costs = {
        "lodgings": sum(graph_data["V"][k].cost for k in lodging_vars.keys()),
        "origins": origin_cost,
        "waypoints": sum(graph_data["V"][k].cost for k in waypoint_vars.keys()),
    }

    print_summary(outputs, counts, costs, calculated_value)
    if "town_1343" in waypoint_vars.keys():
        print("Ancado Inner Harbor active (cost 1) and included with waypoints.")

    return workerman_json



# generate_reference_data.py

import hashlib
import json

import bdo_empire.data_store as ds
from bdo_empire.generate_value_data import generate_value_data


def get_data_files(data: dict) -> None:
    print("Reading data files...")
    data["all_plantzones"] = ds.read_json("plantzone.json")
    data["plantzone_drops"] = ds.read_json("plantzone_drops.json")
    data["lodging_data"] = ds.read_json("all_lodging_storage.json")
    data["town_to_group"] = ds.read_json("town_node_translate.json")["tnk2tk"]
    data["group_to_town"] = ds.read_json("town_node_translate.json")["tk2tnk"]
    data["group_to_townname"] = ds.read_json("warehouse_to_townname.json")
    data["waypoint_data"] = ds.read_json("exploration.json")
    data["waypoint_links"] = ds.read_json("deck_links.json")


def get_value_data(prices: dict, modifiers: dict, data: dict) -> None:
    print("Generating node values...")
    sha_filename = "values_hash.txt"
    current_sha = ds.read_text(sha_filename) if ds.is_file(sha_filename) else None

    encoded = json.dumps({"p": prices, "m": modifiers}).encode()
    latest_sha = hashlib.sha256(encoded).hexdigest()

    if latest_sha == current_sha:
        print("  ...re-using existing node values data.")
    else:
        generate_value_data(prices, modifiers)
        ds.path().joinpath(sha_filename).write_text(latest_sha)

    data["plant_values"] = ds.read_json("node_values_per_town.json")
    data["plants"] = data["plant_values"].keys()
    data["groups"] = data["plant_values"][list(data["plants"])[0]].keys()
    data["towns"] = [data["group_to_town"][w] for w in data["groups"]]
    data["max_ub"] = len(data["plants"])


def get_lodging_data(lodging: dict, data: dict) -> None:
    print("Generating lodging data...")
    for group, lodgings in data["lodging_data"].items():
        if group not in data["groups"]:
            continue
        townname = data["group_to_townname"][group]
        max_lodging = 1 + lodging[townname] + max([int(k) for k in lodgings.keys()])
        data["lodging_data"][group]["max_ub"] = max_lodging
        data["lodging_data"][group]["lodging_bonus"] = lodging[townname]


def generate_reference_data(config: dict, prices: dict, modifiers: dict, lodging: dict) -> dict:
    data = {}
    data["config"] = config
    get_data_files(data)
    get_value_data(prices, modifiers, data)
    get_lodging_data(lodging, data)
    return data



# optimize.py

from pulp import HiGHS, LpVariable, lpSum, LpProblem, LpMaximize

from bdo_empire.generate_graph_data import Arc, GraphData, Node, NodeType as NT
from bdo_empire.optimize_par import solve_par


def filter_arcs(v: Node, groupflow: str, arcs: list[Arc]) -> list[Arc]:
    return [
        var
        for arc in arcs
        for key, var in arc.vars.items()
        if key.startswith("groupflow_") and (key == groupflow or v.isLodging)
    ]


def link_in_out_by_group(prob: LpProblem, v: Node, in_arcs: list[Arc], out_arcs: list[Arc]) -> None:
    all_inflows = []
    f = v.vars["f"]
    for group in v.groups:
        groupflow_key = f"groupflow_{group.id}"
        inflows = filter_arcs(v, groupflow_key, in_arcs)
        outflows = filter_arcs(v, groupflow_key, out_arcs)
        prob += lpSum(inflows) == lpSum(outflows), f"balance_{groupflow_key}_at_{v.name()}"
        all_inflows.append(inflows)
    prob += f == lpSum(all_inflows), f"flow_{v.name()}"
    prob += f <= v.ub * v.vars["x"], f"x_{v.name()}"


def create_problem(config: dict, G: GraphData) -> LpProblem:
    """Create the problem and add the variables and constraints."""

    prob = LpProblem(config["name"], LpMaximize)

    # Variables
    cost = LpVariable("cost", 0, config["budget"], "Integer")

    for v in G["V"].values():
        v.vars["x"] = LpVariable(f"x_{v.name()}", 0, 1, "Binary")
        v.vars["f"] = LpVariable(f"flow_{v.name()}", 0, v.ub, "Integer")

    for arc in G["E"].values():
        for group in set(arc.source.groups).intersection(set(arc.destination.groups)):
            key = f"groupflow_{group.id}"
            ub = arc.ub if arc.source.type in [NT.group, NT.𝓢, NT.𝓣, NT.lodging] else group.ub
            cat = "Binary" if arc.source.type in [NT.𝓢, NT.plant] else "Integer"
            arc.vars[key] = LpVariable(f"{key}_on_{arc.name()}", 0, ub, cat)

    # Objective
    prize_values = [
        round(plant.group_prizes[group.id]["value"], 2) * arc.vars[f"groupflow_{group.id}"]
        for plant in G["P"].values()
        for group in plant.groups
        for arc in plant.inbound_arcs
    ]
    prob += lpSum(prize_values), "ObjectiveFunction"

    # Constraints
    prob += cost == lpSum(v.cost * v.vars["x"] for v in G["V"].values()), "TotalCost"

    for group in G["G"].values():
        vars = [lodge.vars["x"] for lodge in G["L"].values() if lodge.groups[0] == group]
        prob += lpSum(vars) <= 1, f"lodging_{group.id}"

    for v in G["V"].values():
        if v.type not in [NT.𝓢, NT.𝓣]:
            link_in_out_by_group(prob, v, v.inbound_arcs, v.outbound_arcs)

    link_in_out_by_group(prob, G["V"]["𝓣"], G["V"]["𝓣"].inbound_arcs, G["V"]["𝓢"].outbound_arcs)
    prob += G["V"]["𝓢"].vars["x"] == 1, "x_source"

    for node in G["V"].values():
        if node.type in [NT.S, NT.T]:
            continue

        in_neighbors = [arc.source.vars["x"] for arc in node.inbound_arcs]
        out_neighbors = [arc.destination.vars["x"] for arc in node.outbound_arcs]
        if node.isWaypoint:
            prob += lpSum(in_neighbors) - 2 * node.vars["x"] >= 0
        else:
            prob += lpSum(in_neighbors) + lpSum(out_neighbors) - 2 * node.vars["x"] >= 0
        prob += lpSum(out_neighbors) >= node.vars["x"]

    # Edge case handling.
    # If group 619 is active it must be connected to a near town.
    # There are three connection paths to select from...
    connect_sets = [[1321, 1327, 1328, 1329, 1376], [1321, 1327, 1328, 1329, 1330, 1375], [1339]]
    connect_vars = []
    for i, connect_set in enumerate(connect_sets):
        x = LpVariable(f"x_group_619_connect_{i}", 0, 1, "Binary")
        connect_vars.append(x)
        prob += (
            lpSum([G["V"][f"waypoint_{wp}"].vars["x"] for wp in connect_set]) >= len(connect_set) * x
        )
    prob += lpSum(connect_vars) >= G["V"]["group_619"].vars["x"]

    return prob


def optimize(data: dict, graph_data: GraphData) -> LpProblem:
    num_processes = data["config"]["solver"]["num_processes"]
    print(
        f"\nSolving:  graph with {len(graph_data['V'])} nodes and {len(graph_data['E'])} arcs"
        f"\n  Using:  budget of {data["config"]["budget"]}"
        f"\n   With:  {num_processes} processes."
    )

    print("Creating mip problem...")
    prob = create_problem(data["config"], graph_data)
    print("Solving mip problem...")

    options = {k: v for k, v in data["config"]["solver"].items() if k != "num_processes"}

    if num_processes == 1:
        print(f"Single process starting using {options}")
        solver = HiGHS()
        solver.optionsDict = options
        prob.solve(solver)
    else:
        prob = solve_par(prob, options, num_processes)

    return prob



# main.py

from enum import Enum
import json
from math import inf
from pathlib import Path
from random import randint
from tkinter import DISABLED, NORMAL, filedialog

import customtkinter as ctk
from CTkToolTip import CTkToolTip as ctktt
from psutil import cpu_count

from bdo_empire.initialize import initialize_data
from bdo_empire.generate_graph_data import generate_graph_data
from bdo_empire.generate_reference_data import generate_reference_data
from bdo_empire.generate_workerman_data import generate_workerman_data
from bdo_empire.optimize import optimize


solver_config = {
    "num_processes": max(1, cpu_count(logical=False) - 1),
    "mip_rel_gap": 1e-4,
    "mip_feasibility_tolerance": 1e-4,
    "primal_feasibility_tolerance": 1e-4,
    "time_limit": inf,
    "random_seed": randint(0, 2147483647),
}

purchased_lodging = {
    "Velia": 0,
    "Heidel": 0,
    "Glish": 0,
    "Calpheon City": 0,
    "Olvia": 0,
    "Keplan": 0,
    "Port Epheria": 0,
    "Trent": 0,
    "Iliya Island": 0,
    "Altinova": 0,
    "Tarif": 0,
    "Valencia City": 0,
    "Shakatu": 0,
    "Sand Grain Bazaar": 0,
    "Ancado Inner Harbor": 0,
    "Arehaza": 0,
    "Old Wisdom Tree": 0,
    "Grána": 0,
    "Duvencrune": 0,
    "O'draxxia": 0,
    "Eilton": 0,
    "Dalbeol Village": 0,
    "Nampo's Moodle Village": 0,
    "Nopsae's Byeot County": 0,
    "Muzgar": 0,
    "Yukjo Street": 0,
    "Godu Village": 0,
    "Bukpo": 0,
}


class WidgetState(Enum):
    Ready = 0
    Required = 1
    Optional = 2
    Waiting = 3
    Running = 4
    Error = 5


class EmpireOptimizerApp(ctk.CTk):
    def __init__(self):
        super().__init__()

        self.title("Empire Optimizer")
        self.geometry("660x300")

        self.data_state = WidgetState.Required
        self.prices_state = WidgetState.Required
        self.modifiers_state = WidgetState.Optional
        self.lodging_state = WidgetState.Optional
        self.cp_state = WidgetState.Required
        self.outpath_state = WidgetState.Required
        self.optimize_state = WidgetState.Waiting

        self.create_widgets()

    def create_widgets(self):
        row = 0

        row += 1
        self.cp_label = ctk.CTkLabel(self, text="CP Limit")
        self.cp_label.grid(row=row, column=0, padx=10, pady=10)
        self.cp_entry = ctk.CTkEntry(self, validate="focusout", validatecommand=self.validate_cp)
        self.cp_entry.grid(row=row, column=1, padx=10, pady=10)
        self.cp_status = ctk.CTkLabel(self, text=self.cp_state.name)
        self.cp_status.grid(row=row, column=3, padx=10, pady=10)

        row += 1
        self.prices_label = ctk.CTkLabel(self, text="Prices")
        self.prices_label.grid(row=row, column=0, padx=10, pady=10)
        self.prices_entry = ctk.CTkEntry(self, width=300)
        self.prices_entry.grid(row=row, column=1, padx=10, pady=10)
        self.prices_button = ctk.CTkButton(self, text="Browse", command=self.browse_prices_file)
        self.prices_button.grid(row=row, column=2, padx=10, pady=10)
        self.prices_status = ctk.CTkLabel(self, text=self.prices_state.name)
        self.prices_status.grid(row=row, column=3, padx=10, pady=10)
        ctktt(self.prices_label, message="Set to file exported from workerman's settings page.")
        ctktt(self.prices_entry, message="Set to file exported from workerman's settings page.")

        row += 1
        self.lodging_button = ctk.CTkButton(
            self, text="Setup Purchased Lodging", command=self.setup_lodging
        )
        self.lodging_button.grid(row=row, column=1, padx=0, pady=10)
        self.lodging_status = ctk.CTkLabel(self, text=self.lodging_state.name)
        self.lodging_status.grid(row=row, column=3, padx=0, pady=10)
        ctktt(self.lodging_button, message="Setup loyalty and pearl shop lodging.")

        row += 1
        self.modifiers_label = ctk.CTkLabel(self, text="Modifiers")
        self.modifiers_label.grid(row=row, column=0, padx=10, pady=10)
        self.modifiers_entry = ctk.CTkEntry(self, width=300)
        self.modifiers_entry.grid(row=row, column=1, padx=10, pady=10)
        self.modifiers_button = ctk.CTkButton(
            self, text="Browse", command=self.browse_modifiers_file
        )
        self.modifiers_button.grid(row=row, column=2, padx=10, pady=10)
        self.modifiers_status = ctk.CTkLabel(self, text=self.modifiers_state.name)
        self.modifiers_status.grid(row=row, column=3, padx=10, pady=10)
        ctktt(self.modifiers_label, message="Set to file exported from workerman's modifiers page.")
        ctktt(self.modifiers_entry, message="Set to file exported from workerman's modifiers page.")

        row += 1
        self.outpath_label = ctk.CTkLabel(self, text="Output Path")
        self.outpath_label.grid(row=row, column=0, padx=10, pady=10)
        self.outpath_entry = ctk.CTkEntry(self, width=300)
        self.outpath_entry.grid(row=row, column=1, padx=10, pady=10)
        self.outpath_button = ctk.CTkButton(self, text="Browse", command=self.browse_outpath)
        self.outpath_button.grid(row=row, column=2, padx=10, pady=10)
        self.outpath_status = ctk.CTkLabel(self, text=self.outpath_state.name)
        self.outpath_status.grid(row=row, column=3, padx=10, pady=10)

        row += 1
        self.optimize_button = ctk.CTkButton(
            self, text="Optimize", command=self.optimize, state=DISABLED
        )
        self.optimize_button.grid(row=row, column=1, padx=10, pady=10)
        self.optimize_status = ctk.CTkLabel(self, text=self.optimize_state.name)
        self.optimize_status.grid(row=row, column=3, padx=10, pady=10)
        self.config_button = ctk.CTkButton(self, text="Config Solver", command=self.config_solver)
        self.config_button.grid(row=row, column=2, padx=10, pady=10)

    def browse_prices_file(self):
        file_path = filedialog.askopenfilename(filetypes=[("JSON Files", "*.json")])
        if file_path:
            self.prices_entry.delete(0, ctk.END)
            self.prices_entry.insert(0, file_path)
            self.validate_prices(file_path)

    def browse_modifiers_file(self):
        file_path = filedialog.askopenfilename(filetypes=[("JSON Files", "*.json")])
        if file_path:
            self.modifiers_entry.delete(0, ctk.END)
            self.modifiers_entry.insert(0, file_path)
            self.validate_modifiers(file_path)

    def setup_lodging(self):
        global purchased_lodging

        def validate_entry(entry_var, label_widget):
            if entry_var.get().isdigit():
                label_widget.configure(text="Valid", text_color="green")
            else:
                label_widget.configure(text="Invalid", text_color="red")

        lodging_window = ctk.CTkToplevel(self)
        lodging_window.title("Purchased Lodging Setup")
        lodging_window.geometry("430x600")
        lodging_window.grab_set()

        scrollable_frame = ctk.CTkScrollableFrame(lodging_window, width=400, height=500)
        scrollable_frame.grid(row=0, column=0, columnspan=2, padx=10, pady=10)

        # Bind mouse wheel scrolling for linux (WSL)
        scrollable_frame.bind_all(
            "<Button-4>", lambda e: scrollable_frame._parent_canvas.yview("scroll", -1, "units")
        )
        scrollable_frame.bind_all(
            "<Button-5>", lambda e: scrollable_frame._parent_canvas.yview("scroll", 1, "units")
        )

        # Create entries for each town inside the scrollable frame
        self.lodging_entries = {}
        row = 0
        for town, value in purchased_lodging.items():
            label = ctk.CTkLabel(scrollable_frame, text=town)
            label.grid(row=row, column=0, padx=10, pady=5)

            entry_var = ctk.StringVar(value=str(value))
            entry = ctk.CTkEntry(scrollable_frame, textvariable=entry_var)
            entry.grid(row=row, column=1, padx=10, pady=5)

            status_label = ctk.CTkLabel(scrollable_frame, text="Optional")
            status_label.grid(row=row, column=2, padx=10, pady=5)

            entry.bind("<Tab>", lambda e, entry=entry: self.check_scroll(entry, scrollable_frame))

            # I'm not sure if this will work on windows.
            entry.bind(
                "<ISO_Left_Tab>",
                lambda e, entry=entry: self.check_scroll(entry, scrollable_frame),
            )
            # But according to stackoverflow this does.
            entry.bind(
                "<Shift-KeyPress-Tab>",
                lambda e, entry=entry: self.check_scroll(entry, scrollable_frame),
            )

            entry.bind(
                "<FocusOut>", lambda e, var=entry_var, lbl=status_label: validate_entry(var, lbl)
            )

            self.lodging_entries[town] = entry_var
            row += 1

        import_button = ctk.CTkButton(lodging_window, text="Import", command=self.import_lodging)
        import_button.grid(row=1, column=0, padx=10, pady=10)
        export_button = ctk.CTkButton(lodging_window, text="Export", command=self.export_lodging)
        export_button.grid(row=1, column=1, padx=10, pady=10)

        lodging_window.protocol("WM_DELETE_WINDOW", lambda: self.save_lodging_data(lodging_window))

    def check_scroll(self, entry, scrollable_frame, ypad=10, top_buffer=40, bottom_buffer=40):
        entry_row = entry.grid_info()["row"]
        entry_widget = scrollable_frame.grid_slaves(row=entry_row, column=1)[0]

        canvas = scrollable_frame._parent_canvas
        canvas.update_idletasks()
        canvas_height = canvas.winfo_height()
        scroll_top = canvas.canvasy(0)
        scroll_bottom = scroll_top + canvas_height

        widget_y = entry_widget.winfo_y() - ypad
        widget_height = entry_widget.winfo_height() + ypad

        if widget_y < scroll_top + top_buffer:
            target_y = max(0, widget_y - top_buffer)
            canvas.yview_moveto(target_y / scrollable_frame.winfo_height())
        elif (widget_y + widget_height) > scroll_bottom - bottom_buffer:
            target_y = widget_y - (canvas_height - widget_height - bottom_buffer)
            canvas.yview_moveto(target_y / scrollable_frame.winfo_height())

        focused_widget = self.focus_get()
        if focused_widget and focused_widget == entry.tk_focusNext():
            entry.tk_focusNext().focus()
        else:
            entry.tk_focusPrev().focus()

    def save_lodging_data(self, lodging_window):
        global purchased_lodging
        for town, var in self.lodging_entries.items():
            value = var.get()
            purchased_lodging[town] = int(value) if value.isdigit() else 0
        lodging_window.destroy()
        if any(v > 0 for v in purchased_lodging.values()):
            self.lodging_state = WidgetState.Ready
            self.lodging_status.configure(text=self.lodging_state.name, text_color="green")
            self.lodging_status.update()

    def import_lodging(self):
        file_path = filedialog.askopenfilename(filetypes=[("JSON files", "*.json")])
        if file_path:
            with open(file_path, "r") as file:
                loaded_data = json.load(file)
                for town, value in loaded_data.items():
                    if town in self.lodging_entries:
                        self.lodging_entries[town].set(value)

    def export_lodging(self):
        file_path = filedialog.asksaveasfilename(
            defaultextension=".json", filetypes=[("JSON files", "*.json")]
        )
        if file_path:
            lodging_data = {town: var.get() for town, var in self.lodging_entries.items()}
            with open(file_path, "w") as file:
                json.dump(lodging_data, file, indent=4)

    def browse_outpath(self):
        file_path = filedialog.askdirectory()
        if file_path:
            self.outpath_entry.delete(0, ctk.END)
            self.outpath_entry.insert(0, file_path)
            self.validate_outpath(file_path)

    def validate_cp(self):
        entry = self.cp_entry.get()
        if entry.isdigit():
            self.cp_state = WidgetState.Ready
            self.cp_status.configure(text=self.cp_state.name, text_color="green")
            status = True
        elif entry:
            self.cp_state = WidgetState.Error
            self.cp_status.configure(text=self.cp_state.name, text_color="red")
            status = False
        else:
            self.cp_state = WidgetState.Required
            self.cp_status.configure(text=self.cp_state.name)
            status = False
        self.update_optimize_button_state()
        return status

    def validate_prices(self, file_path):
        if Path(file_path).is_file():
            self.prices_state = WidgetState.Ready
            self.prices_status.configure(text=self.prices_state.name, text_color="green")
        else:
            self.prices_state = WidgetState.Error
            self.prices_status.configure(text=self.prices_state.name, text_color="red")
        self.prices_status.update()
        self.update_optimize_button_state()

    def validate_modifiers(self, file_path):
        if Path(file_path).is_file():
            self.modifiers_state = WidgetState.Ready
            self.modifiers_status.configure(text=self.modifiers_state.name, text_color="green")
        else:
            self.modifiers_state = WidgetState.Error
            self.modifiers_status.configure(text=self.modifiers_state.name, text_color="red")
        self.modifiers_status.update()
        self.update_optimize_button_state()

    def validate_outpath(self, file_path):
        if Path(file_path).exists():
            self.outpath_state = WidgetState.Ready
            self.outpath_status.configure(text=self.outpath_state.name, text_color="green")
        else:
            self.outpath_state = WidgetState.Error
            self.outpath_status.configure(text=self.outpath_state.name, text_color="red")
        self.modifiers_status.update()
        self.update_optimize_button_state()

    def update_optimize_button_state(self):
        if self.cp_entry.get().isdigit():
            self.cp_state = WidgetState.Ready

        if (
            self.prices_state is WidgetState.Ready
            and self.modifiers_state is not WidgetState.Error
            and self.lodging_state is not WidgetState.Error
            and self.cp_state is WidgetState.Ready
            and self.outpath_state is WidgetState.Ready
        ):
            self.optimize_state = WidgetState.Ready
            self.optimize_button.configure(state=NORMAL)
        else:
            self.optimize_button.configure(state=DISABLED)
        self.optimize_button.update()

    def config_solver(self):
        global solver_config

        config_window = ctk.CTkToplevel(self)
        config_window.title("Solver Configuration")
        config_window.geometry("400x250")
        config_window.grab_set()

        self.config_entries = {}
        row = 0
        for setting, value in solver_config.items():
            label = ctk.CTkLabel(config_window, text=setting)
            label.grid(row=row, column=0, padx=10, pady=5)

            entry_var = ctk.StringVar(value=str(value))
            entry = ctk.CTkEntry(config_window, textvariable=entry_var)
            entry.grid(row=row, column=1, padx=10, pady=5)

            self.config_entries[setting] = entry_var
            row += 1
        config_window.protocol("WM_DELETE_WINDOW", lambda: self.save_config_data(config_window))

    def save_config_data(self, config_window):
        global solver_config
        int_fields = ["num_processes", "random_seed"]
        for setting, var in self.config_entries.items():
            value = var.get()
            solver_config[setting] = int(value) if setting in int_fields else float(value)
        config_window.destroy()

    def optimize(self):
        global solver_config

        print("Begin optimization...")
        self.optimize_state = WidgetState.Running
        self.optimize_status.configure(text=self.optimize_state.name, text_color="green")
        self.optimize_status.update()

        config = {}
        config["name"] = "Empire"
        config["budget"] = int(self.cp_entry.get())
        config["top_n"] = 4
        config["nearest_n"] = 5
        config["waypoint_ub"] = 25
        config["solver"] = solver_config

        lodging = purchased_lodging
        prices = json.loads(Path(self.prices_entry.get()).read_text())["effectivePrices"]
        modifiers = (
            json.loads(Path(self.modifiers_entry.get()).read_text())["regionModifiers"]
            if self.modifiers_entry.get()
            else {}
        )

        data = generate_reference_data(config, prices, modifiers, lodging)
        graph_data = generate_graph_data(data)
        prob = optimize(data, graph_data)
        workerman_json = generate_workerman_data(prob, lodging, data, graph_data)

        outpath = Path(self.outpath_entry.get())
        outfile = outpath.joinpath("optimized_empire.json")
        with open(outfile, "w") as json_file:
            json.dump(workerman_json, json_file, indent=4)
        print("workerman json written to:", outfile)
        print("Completed.")

        self.optimize_state = WidgetState.Waiting
        self.optimize_status.configure(text=self.optimize_state.name)
        self.optimize_status.update()


def main():
    initialize_data()
    app = EmpireOptimizerApp()
    app.mainloop()


if __name__ == "__main__":
    main()



# data_store.py

import importlib.resources
import json
from pathlib import Path
from urllib import request


def path() -> Path:
    with importlib.resources.as_file(importlib.resources.files().joinpath("data")) as path:
        return path


def is_file(filename: str) -> bool:
    return path().joinpath(filename).is_file()


def read_text(filename: str) -> str:
    return path().joinpath(filename).read_text(encoding="utf-8")


def read_json(filename: str) -> dict:
    content = read_text(filename)
    return json.loads(content)


def write_json(filename: str, data: dict | str) -> None:
    if isinstance(data, str):
        try:
            data = json.loads(data)
        except json.JSONDecodeError:
            raise ValueError("Data is a string but not valid JSON")
    with path().joinpath(filename).open("w", encoding="utf-8") as data_file:
        json.dump(data, data_file, indent=4)


def request_content(url: str) -> str:
    import certifi
    import ssl

    context = ssl.create_default_context(cafile=certifi.where())
    try:
        with request.urlopen(url, context=context) as response:
            content = response.read().decode("utf-8")
    except Exception as e:
        print(f"Error fetching content: {e}")
        raise
    return content


def download_json(filename: str) -> None:
    url = "https://raw.githubusercontent.com/shrddr/workermanjs/refs/heads/main/data"
    if filename in ["plantzone_drops.json", "skills.json"]:
        url = f"{url}/manual/{filename}"
    else:
        url = f"{url}/{filename}"
    content = request_content(url)
    write_json(filename, content)


def download_sha() -> str:
    url = "https://api.github.com/repos/shrddr/workermanjs/branches/main"
    content = request_content(url)
    json_data = json.loads(content)
    return json_data["commit"]["sha"]


def initialized(last_sha: str, filenames: list[str]) -> bool:
    filename = "git_commit.txt"
    current_sha = read_text(filename) if is_file(filename) else None
    return last_sha == current_sha and all(is_file(f) for f in filenames)



# initialize.py

import json
import bdo_empire.data_store as ds


workerman_data_filenames = [
    # Used for node value generation
    "distances_tk2pzk.json",
    "plantzone.json",
    "plantzone_drops.json",
    "skills.json",
    "worker_static.json",
    # Used for empire data generation
    "all_lodging_storage.json",
    "deck_links.json",
    "exploration.json",
    "plantzone.json",
]

local_data_filenames = [
    # Used for empire data generation
    "town_node_translate.json",
    "townnames.json",
    "warehouse_to_townname.json",
]


def extract_tk2tnk_from_js() -> None:
    import re

    url = "https://raw.githubusercontent.com/shrddr/workermanjs/refs/heads/main/src/stores/game.js"
    js_content = ds.request_content(url)

    # re to match the _tk2tnk dictionary
    dict_pattern = re.compile(r"this\._tk2tnk\s*=\s*{([^}]*)}", re.DOTALL)
    match = dict_pattern.search(js_content)
    # Since shrddr's comment in game.js indicates reading this from the bss in the future...
    if not match:
        raise ValueError("tk2tnk dictionary not found, check the game.js source file!")

    dict_content = match.group(1).strip()
    pair_pattern = re.compile(r"(\d+)\s*:\s*(\d+)")
    tk2tnk_dict = {}
    for line in dict_content.splitlines():
        pair_match = pair_pattern.search(line)
        if pair_match:
            key, value = int(pair_match.group(1)), int(pair_match.group(2))
            tk2tnk_dict[key] = str(value)
    tnk2tk_dict = {v: str(k) for k, v in tk2tnk_dict.items()}

    json_data = {"tk2tnk": tk2tnk_dict, "tnk2tk": tnk2tk_dict}
    ds.write_json("town_node_translate.json", json_data)


def extract_town_names() -> dict:
    url = "https://raw.githubusercontent.com/shrddr/workermanjs/refs/heads/main/data/loc.json"
    json_content = ds.request_content(url)
    json_data = json.loads(json_content)
    json_data = json_data["en"]["town"]
    ds.write_json("warehouse_to_townname.json", json_data)
    return json_data


def generate_warehouse_to_town_names(town_names: dict) -> None:
    translator = ds.read_json("town_node_translate.json")
    json_data = {translator["tk2tnk"][k]: v for k, v in town_names.items()}
    ds.write_json("townnames.json", json_data)


def initialize_workerman_data(last_sha: str) -> None:
    for filename in workerman_data_filenames:
        print(f"Getting `{filename}`...", end="")
        ds.download_json(filename)
        print("complete.")

    print("Generating `town_node_translate.json`...", end="")
    extract_tk2tnk_from_js()
    print("complete.")

    print("Extracting town name list...", end="")
    town_names = extract_town_names()
    print("complete.")

    print("Generating warehouse to town name list...", end="")
    generate_warehouse_to_town_names(town_names)
    print("complete.")

    ds.path().joinpath("git_commit.txt").write_text(last_sha)


def initialize_data() -> None:
    # Any error in initialization is fatal and is raised via urllib.
    print("Checking data files...")
    last_sha = ds.download_sha()
    if not ds.initialized(last_sha, workerman_data_filenames + local_data_filenames):
        initialize_workerman_data(last_sha)
    print("Initialized...")



# generate_graph.data.py

from __future__ import annotations
from enum import IntEnum, auto
from typing import Any, Dict, List, TypedDict

import networkx as nx


class GraphData(TypedDict):
    V: Dict[str, Node]
    E: Dict[tuple[str, str], Arc]
    G: Dict[str, Node]
    P: Dict[str, Node]
    L: Dict[str, Node]


class NodeType(IntEnum):
    𝓢 = auto()
    plant = auto()
    waypoint = auto()
    town = auto()
    group = auto()
    lodging = auto()
    𝓣 = auto()

    INVALID = auto()

    def __repr__(self):
        return self.name


class Node:
    def __init__(
        self,
        id: str,
        type: NodeType,
        ub: int,
        lb: int = 0,
        cost: int = 0,
        groups: List[Node] = [],
    ):
        self.id = id
        self.type = type
        self.ub = ub
        self.lb = lb
        self.cost = cost
        self.group_prizes: Dict[str, Dict[str, Any]] = {}
        self.groups = groups if groups else []
        self.key = self.name()
        self.inbound_arcs: List[Arc] = []
        self.outbound_arcs: List[Arc] = []
        self.vars = {}
        self.isPlant = type == NodeType.plant
        self.isLodging = type == NodeType.lodging
        self.isTown = type == NodeType.town
        self.isWaypoint = type == NodeType.waypoint
        self.isGroup = type == NodeType.group

    def name(self) -> str:
        if self.type in [NodeType.𝓢, NodeType.𝓣]:
            return self.id
        return f"{self.type.name}_{self.id}"

    def inSolution(self):
        x_var = self.vars.get("x", None)
        if x_var is not None:
            return x_var.varValue is not None and round(x_var.varValue) >= 1
        else:
            return False

    def as_dict(self) -> Dict[str, Any]:
        obj_dict = {
            "key": self.name(),
            "name": self.name(),
            "id": self.id,
            "type": self.type.name.lower(),
            "ub": self.ub,
            "lb": self.ub,
            "cost": self.cost,
            "group_prizes": self.group_prizes,
            "groups": [],
            "inbound_arcs": [arc.key for arc in self.inbound_arcs],
            "outbound_arcs": [arc.key for arc in self.outbound_arcs],
            "vars": {},
        }
        for node in self.groups:
            if node is self:
                obj_dict["groups"].append("self")
            else:
                obj_dict["groups"].append(node.name())
        for k, v in self.vars.items():
            obj_dict["vars"][k] = v.to_dict()
        return obj_dict

    def __repr__(self) -> str:
        return f"Node(name: {self.name()}, ub: {self.ub}, lb: {self.lb}, cost: {self.cost}, value: {self.group_prizes})"

    def __eq__(self, other) -> bool:
        return self.name() == other.name()

    def __hash__(self) -> int:
        return hash((self.name()))


class Arc:
    def __init__(self, source: Node, destination: Node, ub: int, cost: int = 0):
        self.source = source
        self.destination = destination
        self.ub = ub
        self.cost = cost
        self.key = (source.name(), destination.name())
        self.type = (source.type, destination.type)
        self.vars = {}

    def as_dict(self) -> Dict[str, Any]:
        return {
            "key": self.key,
            "name": self.name(),
            "ub": self.ub,
            "type": self.type,
            "source": self.source.name(),
            "destination": self.destination.name(),
            "vars": {k: v.to_dict() for k, v in self.vars.items() if round(v.varValue) > 0},
        }

    def inSolution(self) -> bool:
        return self.source.inSolution() and self.destination.inSolution()

    def name(self) -> str:
        return f"{self.source.name()}_to_{self.destination.name()}"

    def __repr__(self) -> str:
        return f"arc({self.source.name()} -> {self.destination.name()}, ub: {self.ub})"

    def __eq__(self, other) -> bool:
        return (self.source, self.destination) == (other.source, other.destination)

    def __hash__(self) -> int:
        return hash((self.source.name() + self.destination.name()))


def add_arcs(nodes: Dict[str, Node], arcs: Dict[tuple, Arc], node_a: Node, node_b: Node):
    """Add arcs between a and b."""
    # A safety measure to ensure arc direction.
    if node_a.type > node_b.type:
        node_a, node_b = node_b, node_a

    arc_configurations = {
        (NodeType.𝓢, NodeType.plant): (1, 0),
        (NodeType.plant, NodeType.waypoint): (1, 0),
        (NodeType.plant, NodeType.town): (1, 0),
        (NodeType.waypoint, NodeType.waypoint): (node_b.ub, node_a.ub),
        (NodeType.waypoint, NodeType.town): (node_b.ub, node_a.ub),
        (NodeType.town, NodeType.town): (node_b.ub, node_a.ub),
        (NodeType.town, NodeType.group): (node_b.ub, 0),
        (NodeType.group, NodeType.lodging): (node_b.ub, 0),
        (NodeType.lodging, NodeType.𝓣): (node_a.ub, 0),
    }

    ub, reverse_ub = arc_configurations.get((node_a.type, node_b.type), (1, 0))

    arc_a = Arc(node_a, node_b, ub=ub)
    arc_b = Arc(node_b, node_a, ub=reverse_ub)

    for arc in [arc_a, arc_b]:
        if arc.key not in arcs and arc.ub > 0:
            arcs[arc.key] = arc
            nodes[arc.source.key].outbound_arcs.append(arc)
            nodes[arc.destination.key].inbound_arcs.append(arc)

            if arc.destination.type is NodeType.lodging:
                arc.destination.groups = [arc.source]


def get_sparsified_link_graph(ref_data: Dict[str, Any]):
    link_graph = nx.Graph()
    for link in ref_data["waypoint_links"]:
        link_graph.add_edge(link[0], link[1])
    for node, data in link_graph.nodes(data=True):
        data["weight"] = ref_data["waypoint_data"][str(node)]["CP"]
        data["type"] = get_link_node_type(str(node), ref_data)

    # This just removes the non-plant leaf nodes. Doing any other reductions reduces performance.
    removal_nodes = []
    for node, data in link_graph.nodes(data=True):
        if nx.degree(link_graph, node) == 1 and data["type"] is not NodeType.plant:
            removal_nodes.append(node)
    if removal_nodes:
        link_graph.remove_nodes_from(removal_nodes)
        removal_nodes = []

    return link_graph


def get_link_node_type(node_id: str, ref_data: Dict[str, Any]):
    """Return the NodeType of the given node_id node.

    - NodeType.INVALID indicates a node that is unused and not added to the graph.
    """
    if node_id in ref_data["towns"]:
        return NodeType.town
    if node_id in ref_data["all_plantzones"]:
        if node_id not in ref_data["plants"]:
            return NodeType.INVALID
        return NodeType.plant
    return NodeType.waypoint


def get_link_nodes(nodes, link, ref_data, link_graph):
    if not link_graph.has_node(link[0]) or not link_graph.has_node(link[1]):
        return (None, None)

    node_a_id, node_b_id = str(link[1]), str(link[0])
    node_a_type = get_link_node_type(node_a_id, ref_data)
    node_b_type = get_link_node_type(node_b_id, ref_data)

    # Ensure arc node order.
    if node_a_type > node_b_type:
        node_a_id, node_b_id = node_b_id, node_a_id
        node_a_type, node_b_type = node_b_type, node_a_type

    if NodeType.INVALID in [node_a_type, node_b_type]:
        # Not used in the graph because they are beyond the scope of the optimization.
        return (None, None)

    return (
        get_node(nodes, node_a_id, node_a_type, ref_data),
        get_node(nodes, node_b_id, node_b_type, ref_data),
    )


def get_node(nodes, node_id: str, node_type: NodeType, ref_data: Dict[str, Any], **kwargs) -> Node:
    """
    Generate, add and return node based on NodeType.

    kwargs `plant` and `group` are required for supply nodes.
    kwargs `ub` is required for group nodes.
    kwargs `ub`, `cost` and `group` are required for lodging nodes.
    """

    groups = []
    lb = 0

    match node_type:
        case NodeType.𝓢:
            ub = ref_data["max_ub"]
            cost = 0
        case NodeType.plant:
            ub = 1
            cost = ref_data["waypoint_data"][node_id]["CP"]
        case NodeType.waypoint | NodeType.town:
            ub = ref_data["config"]["waypoint_ub"]
            cost = ref_data["waypoint_data"][node_id]["CP"]
        case NodeType.group:
            lodging_data = ref_data["lodging_data"][node_id]
            ub = lodging_data["max_ub"] + lodging_data["lodging_bonus"]
            ub = min(ub, ref_data["config"]["waypoint_ub"])
            cost = 0
        case NodeType.lodging:
            ub = kwargs.get("ub")
            lb = kwargs.get("lb")
            root = kwargs.get("root")
            cost = kwargs.get("cost")
            assert (
                ub and (lb is not None) and (cost is not None) and root
            ), "Lodging nodes require 'ub', 'lb' 'cost' and 'root' kwargs."
            groups = [root]
        case NodeType.𝓣:
            ub = ref_data["max_ub"]
            cost = 0
        case NodeType.INVALID:
            assert node_type is not NodeType.INVALID, "INVALID node type."
            return  # Unreachable: Stops pyright unbound error reporting.

    node = Node(node_id, node_type, ub, lb, cost, groups)
    if node.key not in nodes:
        if node.type is NodeType.group:
            node.groups = [node]
        nodes[node.key] = node

    return nodes[node.key]


def process_links(nodes: Dict[str, Node], arcs: Dict[tuple, Arc], ref_data: Dict[str, Any]):
    """Process all waypoint links and add the nodes and arcs to the graph.

    Calls handlers for plant and town nodes to add plant value nodes and
    group/lodging nodes with their respective source and sink arcs.
    """
    link_graph = get_sparsified_link_graph(ref_data)

    for link in ref_data["waypoint_links"]:
        source, destination = get_link_nodes(nodes, link, ref_data, link_graph)
        if source is None or destination is None:
            continue

        add_arcs(nodes, arcs, source, destination)

        if source.isPlant:
            process_plant(nodes, arcs, source, ref_data)
        if destination.isTown:
            process_town(nodes, arcs, destination, ref_data)


def process_plant(
    nodes: Dict[str, Node], arcs: Dict[tuple, Arc], plant: Node, ref_data: Dict[str, Any]
):
    """Add plant group values and arcs between the source and plant nodes."""
    for i, (group_id, value_data) in enumerate(ref_data["plant_values"][plant.id].items(), 1):
        if i > ref_data["config"]["top_n"]:
            break
        if value_data["value"] == 0:
            continue
        value_data["value"]
        plant.group_prizes[group_id] = value_data

    add_arcs(nodes, arcs, nodes["𝓢"], plant)


def process_town(
    nodes: Dict[str, Node], arcs: Dict[tuple, Arc], town: Node, ref_data: Dict[str, Any]
):
    """Add town group and lodging nodes and arcs between the town and sink nodes."""
    group_id = ref_data["town_to_group"][town.id]
    lodging_data = ref_data["lodging_data"][group_id]
    lodging_bonus = lodging_data["lodging_bonus"]

    lodgings = [(1 + lodging_bonus, 0)]
    for ub, lodging_data in lodging_data.items():
        if ub in ["max_ub", "lodging_bonus"]:
            continue
        current = (1 + lodging_bonus + int(ub), lodging_data[0].get("cost"))
        while lodgings and current[1] <= lodgings[-1][1] and current[0] >= lodgings[-1][0]:
            lodgings.pop(-1)
        lodgings.append(current)
        if current[0] + 1 >= ref_data["config"]["waypoint_ub"]:
            break

    group_node = get_node(nodes, group_id, NodeType.group, ref_data, ub=lodgings[-1][0])
    add_arcs(nodes, arcs, town, group_node)

    lb = 0
    for ub, cost in lodgings:
        lodging_node = get_node(
            nodes,
            f"{group_node.id}_for_{ub}",
            NodeType.lodging,
            ref_data,
            ub=ub,
            lb=lb,
            cost=cost,
            root=group_node,
        )
        add_arcs(nodes, arcs, group_node, lodging_node)
        add_arcs(nodes, arcs, lodging_node, nodes["𝓣"])
        lb = ub + 1


def nearest_n_towns(ref_data: Dict[str, Any], G: GraphData, nearest_n: int):
    waypoint_graph = nx.DiGraph()
    for arc in G["E"].values():
        weight = 999999 if "1727" in arc.name() else arc.destination.cost
        waypoint_graph.add_edge(arc.source.id, arc.destination.id, weight=weight)
    all_pairs = dict(nx.all_pairs_bellman_ford_path_length(waypoint_graph, weight="weight"))

    nearest_towns_dist = {}
    nearest_towns = {}

    for node_id, node in G["V"].items():
        if node.isWaypoint or node.isTown:
            distances = []
            for group in G["G"].values():
                town_id = ref_data["group_to_town"][group.id]
                distances.append((group, all_pairs[node.id][town_id]))
            nearest_towns_dist[node_id] = sorted(distances, key=lambda x: x[1])[:nearest_n]
            nearest_towns[node_id] = [w for w, _ in nearest_towns_dist[node_id]]

    return nearest_towns


def finalize_groups(ref_data: Dict[str, Any], G: GraphData, nearest_n: int):
    # All group nodes have now been generated, finalize groups entries
    nearest_towns = nearest_n_towns(ref_data, G, nearest_n)
    for v in G["V"].values():
        if v.type in [NodeType.𝓢, NodeType.𝓣]:
            v.groups = [w for w in G["G"].values()]
        elif v.isWaypoint or v.isTown:
            v.groups = [w for w in nearest_towns[v.key]]
        elif v.isPlant:
            v.groups = [w for w in G["G"].values() if w.id in v.group_prizes.keys()]


def generate_graph_data(ref_data):
    """Generate and return a GraphData Dict composing the LP empire data."""
    print("Generating graph data...")
    nodes: Dict[str, Node] = {}
    arcs: Dict[tuple[str, str], Arc] = {}

    get_node(nodes, "𝓢", NodeType.𝓢, ref_data)
    get_node(nodes, "𝓣", NodeType.𝓣, ref_data)
    process_links(nodes, arcs, ref_data)

    G: GraphData = {
        "V": dict(sorted(nodes.items(), key=lambda item: item[1].type)),
        "E": dict(sorted(arcs.items(), key=lambda item: item[1].as_dict()["type"])),
        "G": {k: v for k, v in nodes.items() if v.isGroup},
        "P": {k: v for k, v in nodes.items() if v.isPlant},
        "L": {k: v for k, v in nodes.items() if v.isLodging},
    }
    finalize_groups(ref_data, G, ref_data["config"]["nearest_n"])

    return G



import multiprocessing
from multiprocessing import Queue
from random import randint
from pulp import HiGHS, LpProblem


def solve_par_worker(
    prob: LpProblem, options_dict: dict, queue: Queue, results: list, process_index: int
) -> None:
    options_dict["random_seed"] = randint(0, 2147483647)
    print(f"Process {process_index} starting using {options_dict}")
    solver = HiGHS()
    solver.optionsDict = options_dict
    prob.solve(solver)
    results[process_index] = prob.to_dict()
    queue.put(process_index)
    return


def solve_par(prob: LpProblem, options_dict: dict, num_processes: int) -> LpProblem:
    manager = multiprocessing.Manager()
    processes = []
    queue = multiprocessing.Queue()
    results = manager.list(range(num_processes))

    for i in range(num_processes):
        p = multiprocessing.Process(
            target=solve_par_worker, args=(prob, options_dict, queue, results, i)
        )
        processes.append(p)
        p.start()

    first_process = queue.get()
    for i, process in enumerate(processes):
        if process.is_alive():
            print(f"Terminating process: {i}")
            process.terminate()
        process.join()
    print(f"Using results from process {first_process}")
    result = prob.from_dict(results[first_process])
    return result[1]



# generate_value_data.py

from math import ceil
import bdo_empire.data_store as ds


def get_data_files(data: dict) -> None:
    data["plantzone"] = ds.read_json("plantzone.json")
    data["plantzone_drops"] = ds.read_json("plantzone_drops.json")
    data["worker_skills"] = ds.read_json("skills.json")
    data["worker_static"] = ds.read_json("worker_static.json")
    data["distances_tk2pzk"] = ds.read_json("distances_tk2pzk.json")


def isGiant(charkey: int, data: dict) -> bool:
    return data["worker_static"][str(charkey)]["species"] in [2, 4, 8]


def skill_bonus(skill_set: list, data: dict) -> dict:
    bonus = {"wspd": 0, "mspd": 0, "luck": 0}
    for sk in skill_set:
        skill_bonuses = data["worker_skills"].get(sk, {})
        bonus["wspd"] += skill_bonuses.get("wspd", 0)
        bonus["wspd"] += skill_bonuses.get("wspd_farm", 0)
        bonus["mspd"] += skill_bonuses.get("mspd", 0)
        bonus["luck"] += skill_bonuses.get("luck", 0)
    return bonus


def worker_stats(worker: dict, skill_set, data: dict) -> dict:
    bonus = skill_bonus(skill_set, data)
    wspd = worker["wspd"] + bonus["wspd"]
    mspd_base = data["worker_static"][str(worker["charkey"])]["mspd"] / 100
    mspd = mspd_base * ((worker["mspd"] / mspd_base) + bonus["mspd"] / 100)
    luck = worker["luck"] + bonus["luck"]
    return {"wspd": wspd, "mspd": mspd, "luck": luck}


def calcCyclesDaily(
    baseWorkload: float, wspd: float, dist: float, mspd: float, modifier: float
) -> float:
    moveMinutes = 2 * dist / mspd / 60
    activeWorkload = baseWorkload * (2 - modifier / 100)
    workMinutes = ceil(activeWorkload / wspd)
    cycleMinutes = 10 * workMinutes + moveMinutes
    return 24 * 60 / cycleMinutes


def price_bunch(bunch: dict, data: dict) -> float:
    return sum(data["market_value"][k] * q for k, q in bunch.items())


def price_pzd(pzd: dict, luck: float, data: dict) -> float:
    unlucky_price = price_bunch(pzd.get("unlucky", {}), data)
    if "lucky" in pzd:
        lucky_price = price_bunch(pzd["lucky"], data)
        return (luck / 100) * lucky_price + (1 - luck / 100) * unlucky_price
    return unlucky_price


def price_lerp(lucky_price: float, unlucky_price: float, luck: float) -> float:
    return (luck / 100) * lucky_price + (1 - luck / 100) * unlucky_price


def profitPzTownStats(
    pzk: int, _tnk, dist: float, wspd: float, mspd: float, luck: float, is_giant: bool, data: dict
) -> float:
    if dist == 9999999:
        return 0

    drop = data["plantzone_drops"][str(pzk)]
    luckyPart = price_bunch(drop["lucky"], data)
    unluckyValue = price_bunch(drop["unlucky"], data)
    luckyValue = unluckyValue + luckyPart
    unluckyValue_gi = price_bunch(drop["unlucky_gi"], data)
    luckyValue_gi = unluckyValue_gi + luckyPart

    rgk = data["plantzone"][str(pzk)]["regiongroup"]
    modifier = data["modifiers"].get(str(rgk), 0)
    if modifier == "":
        modifier = 0

    cycleValue = (
        price_lerp(luckyValue_gi, unluckyValue_gi, luck)
        if is_giant
        else price_lerp(luckyValue, unluckyValue, luck)
    )
    cyclesDaily = calcCyclesDaily(drop["workload"], wspd, dist, mspd, modifier)
    priceDaily = cyclesDaily * cycleValue
    return priceDaily


def profit(
    town: str, plantzone: int, dist: float, worker: dict, skill_set: list, data: dict
) -> float:
    stats = worker_stats(worker, skill_set, data)
    priceDaily = profitPzTownStats(
        plantzone,
        town,
        dist,
        stats["wspd"],
        stats["mspd"],
        stats["luck"],
        isGiant(worker["charkey"], data),
        data,
    )
    return priceDaily


def makeMedianChar(charkey: int, data: dict) -> dict:
    stat = data["worker_static"][str(charkey)]
    pa_wspd = stat["wspd"]
    pa_mspdBonus = 0
    pa_luck = stat["luck"]

    for _ in range(2, 41):
        pa_wspd += (stat["wspd_lo"] + stat["wspd_hi"]) / 2
        pa_mspdBonus += (stat["mspd_lo"] + stat["mspd_hi"]) / 2
        pa_luck += (stat["luck_lo"] + stat["luck_hi"]) / 2
    pa_mspd = stat["mspd"] * (1 + pa_mspdBonus / 1e6)

    return {
        "wspd": round(pa_wspd / 1e6 * 100) / 100,
        "mspd": round(pa_mspd) / 100,
        "luck": round(pa_luck / 1e4 * 100) / 100,
        "charkey": charkey,
        "isGiant": isGiant(charkey, data),
    }


def medianGoblin(tnk: int, data: dict) -> dict:
    if tnk == 1623:
        return makeMedianChar(8003, data)  # grana
    if tnk == 1604:
        return makeMedianChar(8003, data)  # owt
    if tnk == 1691:
        return makeMedianChar(8023, data)  # oddy
    if tnk == 1750:
        return makeMedianChar(8035, data)  # eilton
    if tnk == 1781:
        return makeMedianChar(8050, data)  # lotml
    if tnk == 1785:
        return makeMedianChar(8050, data)  # lotml
    if tnk == 1795:
        return makeMedianChar(8050, data)  # lotml
    if tnk == 1857:
        return makeMedianChar(8050, data)  # lotml2
    if tnk == 1858:
        return makeMedianChar(8050, data)  # lotml2
    if tnk == 1853:
        return makeMedianChar(8050, data)  # lotml2
    return makeMedianChar(7572, data)


def medianGiant(tnk: int, data: dict) -> dict:
    if tnk == 1623:
        return makeMedianChar(8006, data)  # grana
    if tnk == 1604:
        return makeMedianChar(8006, data)  # owt
    if tnk == 1691:
        return makeMedianChar(8027, data)  # oddy
    if tnk == 1750:
        return makeMedianChar(8039, data)  # eilton
    if tnk == 1781:
        return makeMedianChar(8058, data)  # lotml
    if tnk == 1785:
        return makeMedianChar(8058, data)  # lotml
    if tnk == 1795:
        return makeMedianChar(8058, data)  # lotml
    if tnk == 1857:
        return makeMedianChar(8058, data)  # lotml2
    if tnk == 1858:
        return makeMedianChar(8058, data)  # lotml2
    if tnk == 1853:
        return makeMedianChar(8058, data)  # lotml2
    return makeMedianChar(7571, data)


def medianHuman(tnk: int, data: dict) -> dict:
    if tnk == 1623:
        return makeMedianChar(8009, data)  # grana
    if tnk == 1604:
        return makeMedianChar(8009, data)  # owt
    if tnk == 1691:
        return makeMedianChar(8031, data)  # oddy
    if tnk == 1750:
        return makeMedianChar(8043, data)  # eilton
    if tnk == 1781:
        return makeMedianChar(8054, data)  # lotml
    if tnk == 1785:
        return makeMedianChar(8054, data)  # lotml
    if tnk == 1795:
        return makeMedianChar(8054, data)  # lotml
    if tnk == 1857:
        return makeMedianChar(8054, data)  # lotml2
    if tnk == 1858:
        return makeMedianChar(8054, data)  # lotml2
    if tnk == 1853:
        return makeMedianChar(8054, data)  # lotml2
    return makeMedianChar(7573, data)


def optimize_skills(town: str, plantzone: int, dist: float, worker: dict, data: dict):
    max_skills = 9
    w_bonuses = {0: {"skills": [], "profit": 0}}
    w_actions = ["wspd"]
    w_actions.append("wspd_farm")

    w_skills = []
    for key, skill in data["worker_skills"].items():
        if any(act in skill for act in w_actions):
            w_skills.append(
                {
                    "key": key,
                    "amount": skill.get("wspd", 0) + skill.get("wspd_farm", 0),
                    "mspd": skill.get("mspd", 0),
                }
            )

    w_skills.sort(key=lambda x: (x["amount"], x["mspd"]), reverse=True)

    for i in range(1, max_skills + 1):
        temp_skills = [w["key"] for w in w_skills[:i]]
        new_profit = profit(town, plantzone, dist, worker, temp_skills, data)
        w_bonuses[i] = {"skills": temp_skills, "profit": new_profit}

        if all(not data["worker_skills"][sk].get("mspd", 0) for sk in temp_skills):
            mod_skills = temp_skills.copy()
            wm_skills = [ss for ss in w_skills if ss["mspd"] > 0]
            if wm_skills:
                mod_skills[-1] = wm_skills[0]["key"]
                mod_profit = profit(town, plantzone, dist, worker, mod_skills, data)
                if mod_profit > new_profit:
                    w_bonuses[i] = {"skills": mod_skills, "profit": mod_profit}

    ml_actions = ["mspd", "luck"]
    ml_skills = {
        key
        for key, skill in data["worker_skills"].items()
        if any(act in skill for act in ml_actions)
    }

    step_results = [w_bonuses[max_skills]]
    ml_best_skills = []
    for i in range(1, max_skills + 1):
        step_base_skills = w_bonuses[max_skills - i]["skills"] + ml_best_skills
        step_candidates = []

        for sk in ml_skills:
            if sk in w_bonuses[max_skills - i]["skills"]:
                continue
            temp_skills = step_base_skills + [sk]
            new_profit = profit(town, plantzone, dist, worker, temp_skills, data)
            step_candidates.append({"sk": sk, "profit": new_profit})

        if step_candidates:
            step_candidates.sort(key=lambda x: x["profit"], reverse=True)
            step_best_skill = step_candidates[0]["sk"]
            step_skills = step_base_skills + [step_best_skill]
            step_results.append({"skills": step_skills, "profit": step_candidates[0]["profit"]})
            ml_best_skills.append(step_best_skill)
            ml_skills.remove(step_best_skill)
        else:
            ml_best_skills.append(0)

    step_results.sort(key=lambda x: x["profit"], reverse=True)
    return step_results[0]


def generate_value_data(prices: dict, modifiers: dict) -> None:
    data = {}
    get_data_files(data)
    data["market_value"] = prices
    data["modifiers"] = modifiers

    # Workerman sorts by nearest node to town.
    for town in data["distances_tk2pzk"]:
        data["distances_tk2pzk"][town] = sorted(data["distances_tk2pzk"][town], key=lambda x: x[0])

    output = {}
    for town in data["distances_tk2pzk"].keys():
        if town in ["1375"]:
            continue

        median_workers = {
            "giant": medianGiant(town, data),
            "goblin": medianGoblin(town, data),
            "human": medianHuman(town, data),
        }

        for dist_data in data["distances_tk2pzk"][town]:
            plantzone, dist = dist_data
            if not data["plantzone"][str(plantzone)]["node"]["is_plantzone"]:
                continue
            if data["plantzone"][str(plantzone)]["node"]["kind"] in [12, 13]:
                continue

            optimized_workers = {
                "giant": optimize_skills(town, plantzone, dist, median_workers["giant"], data),
                "goblin": optimize_skills(town, plantzone, dist, median_workers["goblin"], data),
                "human": optimize_skills(town, plantzone, dist, median_workers["human"], data),
            }
            optimized_worker = max(optimized_workers.items(), key=lambda item: item[1]["profit"])

            if str(plantzone) not in output:
                output[str(plantzone)] = {}
            output[str(plantzone)][str(town)] = {}
            output[str(plantzone)][str(town)]["worker"] = optimized_worker[0]
            output[str(plantzone)][str(town)]["value"] = optimized_worker[1]["profit"]
            output[str(plantzone)][str(town)]["worker_data"] = median_workers[
                optimized_worker[0]
            ].copy()
            output[str(plantzone)][str(town)]["worker_data"]["skills"] = [
                int(s) for s in optimized_worker[1]["skills"].copy()
            ]

    # Make the list a plantzone keyed list sorted by value in descending order for 'top_n'
    for plantzone, warehouse_data in output.copy().items():
        output[plantzone] = dict(
            sorted(warehouse_data.items(), key=lambda x: x[1]["value"], reverse=True)
        )

    ds.write_json("node_values_per_town.json", output)
